import numpy
from dataclasses import dataclass
from .define import Range, ChunkData
from ..internal.symbol_export_timeline_db import release_chunk_timeline
from ..internal.symbol_export_chunk_timeline import (
    ctl_all_time_point,
    ctl_all_time_point_len,
    ctl_append_disk_chunk,
    ctl_append_network_chunk,
    ctl_compact,
    ctl_empty,
    ctl_jump_to_disk_chunk,
    ctl_jump_to_network_chunk,
    ctl_last_disk_chunk,
    ctl_last_network_chunk,
    ctl_next_disk_chunk,
    ctl_next_network_chunk,
    ctl_pointer,
    ctl_pop,
    ctl_read_only,
    ctl_reset_pointer,
    ctl_save,
    ctl_set_max_limit,
)


@dataclass
class ChunkTimeline:
    """
    ChunkTimeline records the timeline of a chunk,
    and it contains the change logs about this chunk
    on this timeline.

    In other words, the ChunkTimeline holds the history
    of this chunk.

    Note that it's unsafe for multiple thread to access this
    struct due to we don't use mutex to ensure the operation
    is atomic.

    So, it's your responsibility to make ensure there is only
    one thread is using this object.

    Additionally, before you use this object, please ensure
    you use ChunkTimeline.is_valid() to check whether the timeline
    is valid or not.
    Note that check only need do just once, and after you do the
    check, you don't need to do more checks when you use the same object.
    """

    _chunk_timeline_id: int = -1

    def __del__(self):
        if self._chunk_timeline_id >= 0 and release_chunk_timeline is not None:
            release_chunk_timeline(self._chunk_timeline_id)

    def is_valid(self) -> bool:
        """
        is_valid check current chunk timeline is valid or not.

        If not valid, it means the chunk timeline actually not exist,
        not only Python but also in Go.

        Try to use an invalid chunk timeline is not allowed,
        and any operation will be terminated.

        Returns:
            bool: Whether the chunk timeline is valid or not.
        """
        return self._chunk_timeline_id >= 0

    def append_disk_chunk(
        self, chunk_data: ChunkData, nop_when_no_change: bool = False
    ):
        """
        append_disk_chunk tries append a new chunk whose is disk
        encoding to the timeline of current chunk. Additionally,
        we will also append the block entities of this chunk.

        Calling append_disk_chunk will make sure there is exist at
        least one empty space to place the new time point, whether
        new time point will be added in the end or not.

        The way to leave empty space is by calling pop, and the poped
        time points must be the most earliest one.

        If current timeline is read only, then calling append_disk_chunk
        will do no operation.

        Args:
            chunk_data (ChunkData): The chunk you want to append to the timeline.
            nop_when_no_change (bool, optional):
                Specific if the append one have no difference between the latest one,
                then don't append anything to the current chunk timeline.
                Defaults to False.


        Raises:
            Exception: When failed to append the chunk.
        """
        err = ctl_append_disk_chunk(
            self._chunk_timeline_id,
            chunk_data.sub_chunks,
            chunk_data.nbts,
            chunk_data.chunk_range.start_range,
            chunk_data.chunk_range.end_range,
            nop_when_no_change,
        )
        if len(err) > 0:
            raise Exception(err)

    def append_network_chunk(
        self, chunk_data: ChunkData, nop_when_no_change: bool = False
    ):
        """
        append_network_chunk tries append a new chunk whose is network
        encoding to the timeline of current chunk.
        Additionally, we will also append the block entities of this chunk.

        Calling append_network_chunk will make sure there is exist at least
        one empty space to place the new time point, whether new time point
        will be added in the end or not.

        The way to leave empty space is by calling pop, and the poped time
        points must be the most earliest one.

        If current timeline is read only, then calling append_network_chunk
        will do no operation.

        Args:
            chunk_data (ChunkData): The chunk you want to append to the timeline.
            nop_when_no_change (bool, optional):
                Specific if the append one have no difference between the latest one,
                then don't append anything to the current chunk timeline.
                Defaults to False.

        Raises:
            Exception: When failed to append the chunk.
        """
        err = ctl_append_network_chunk(
            self._chunk_timeline_id,
            chunk_data.sub_chunks,
            chunk_data.nbts,
            chunk_data.chunk_range.start_range,
            chunk_data.chunk_range.end_range,
            nop_when_no_change,
        )
        if len(err) > 0:
            raise Exception(err)

    def empty(self) -> bool:
        """
        empty returns whether this timeline is empty or not.
        If is empty, then calling Save will result in no operation.

        Returns:
            bool: Return True for exist.
                  Return False for not exist or current timeline is not exist.
        """
        result = ctl_empty(self._chunk_timeline_id)
        return result == 1

    def read_only(self) -> bool:
        """
        read_only returns whether this timeline is read only or not.
        If is read only, then calling any function that will modify
        underlying timeline will result in no operation.

        Returns:
            bool: Return True for this timeline is read only.
                  Return False for this timeline could be modified,
                  or this timeline is not exist.
        """
        result = ctl_read_only(self._chunk_timeline_id)
        return result == 1

    def pointer(self) -> int:
        """pointer returns the index of the next time point that will be read.

        Returns:
            int: The index of the next time point.
                 Return -1 for this timeline is not exist.
        """
        return ctl_pointer(self._chunk_timeline_id)

    def reset_pointer(self):
        """
        reset_pointer resets the pointer to the first time point of this timeline.
        reset_pointer is always successful if there even have no time point.

        Raises:
            Exception: When this timeline is not exist.
        """
        err = ctl_reset_pointer(self._chunk_timeline_id)
        if len(err) > 0:
            raise Exception(err)

    def all_time_point(self) -> numpy.ndarray:
        """
        all_time_point returns a list that holds
        the unix time of all time points this timeline
        have.

        Granted the returned array is non-decreasing.

        Note that the retuened list is read only.
        If need to modify, please copy a new one.

        Returns:
            numpy.ndarray: The list that holds the unix time
                           for all time points in this timeline.
                           If is empty, then this timeline have
                           no time point or this timeline is not
                           exist.
        """
        return ctl_all_time_point(self._chunk_timeline_id)

    def all_time_point_len(self) -> int:
        """
        all_time_point_len returns the length of
        the time point that this timeline have.

        Returns:
            int: The length of this timeline.
        """
        return ctl_all_time_point_len(self._chunk_timeline_id)

    def set_max_limit(self, max_limit: int):
        """
        set_max_limit sets the timeline could record how many time point.
        max_limit must bigger than 0. If less, then set the limit to 1.

        After calling set_max_limit if overflow immediately, then we will
        pop some time point from the underlying timeline.
        Poped time points must be the most earliest one.

        Note that calling set_max_limit will not change the empty states
        of this timeline.

        If current timeline is read only, then calling set_max_limit will
        do no operation.

        Args:
            max_limit (int): The max limit of this timeline.

        Raises:
            Exception: When failed to update the max limit.
        """
        err = ctl_set_max_limit(self._chunk_timeline_id, max_limit)
        if len(err) > 0:
            raise Exception(err)

    def compact(self):
        """
        compact compacts the underlying block palette as much as possible, try to delete all
        unused blocks from it.

        If current timeline is empty or read only, then calling compact will do no operation.

        Note that if you got exception from compact, then the underlying pointer will back to
        the firest time point due to when an error occurs, some of the underlying data maybe is
        inconsistent.

        compact is very expensive due to its time complexity is O(C×k×4096×N×L).
            - k is the count of sub chunks that this chunk have.
            - N is the count of time point that this timeline have.
            - L is the average count of layers for each sub chunks in this timeline.
            - C is a little big (bigger than 2) due to there are multiple operations need to do.

        Raises:
            Exception: When failed to compact the underlying block palette.
        """
        err = ctl_compact(self._chunk_timeline_id)
        if len(err) > 0:
            raise Exception(err)

    def next_disk_chunk(self) -> tuple[ChunkData, int, bool] | None:
        """
        next_disk_chunk gets the next time point of current chunk and the NBT blocks in it.
        Note that the returned ChunkData is in disk encoding.

        With the call to next_disk_chunk, we granted that the returned time keeps increasing
        until the entire time series is traversed.

        When it is already at the end of the timeline, calling next_disk_chunk again will back
        to the earliest time point.
        In other words, next_disk_chunk is self-loop and can be called continuously.

        Note that if return None (meet error), then the underlying pointer will back to the
        firest time point due to when an error occurs, some of the underlying data maybe is
        inconsistent.

        Time complexity: O(4096×n + C).
        n is the sub chunk count of this chunk.
        C is relevant to the average changes between last time point and the next one.

        Returns:
            tuple[ChunkData, int, bool] | None:
                The chunk data in the next time point of this current chunk timeline.
                Returned int is the update unix time of this time point.
                The returned bool can inform whether the element obtained after the
                current call to next_disk_chunk is at the end of the time series.
                If meet error, then return None.
        """
        (
            sub_chunks,
            range_start,
            range_end,
            nbts,
            update_unix_time,
            is_last_element,
            success,
        ) = ctl_next_disk_chunk(self._chunk_timeline_id)

        if not success:
            return None

        return (
            ChunkData(sub_chunks, nbts, Range(range_start, range_end)),
            update_unix_time,
            is_last_element,
        )

    def next_network_chunk(self) -> tuple[ChunkData, int, bool] | None:
        """
        next_network_chunk gets the next time point of current chunk and the NBT blocks in it.
        Note that the returned ChunkData is in network encoding.

        With the call to next_network_chunk, we granted that the returned time keeps increasing
        until the entire time series is traversed.

        When it is already at the end of the timeline, calling next_network_chunk again will back
        to the earliest time point.
        In other words, next_network_chunk is self-loop and can be called continuously.

        Note that if return None (meet error), then the underlying pointer will back to the firest
        time point due to when an error occurs, some of the underlying data maybe is inconsistent.

        Time complexity: O(4096×n + C).
        n is the sub chunk count of this chunk.
        C is relevant to the average changes between last time point and the next one.

        Returns:
            tuple[ChunkData, int, bool] | None:
                The chunk data in the next time point of this current chunk timeline.
                Returned int is the update unix time of this time point.
                The returned bool can inform whether the element obtained after the
                current call to next_network_chunk is at the end of the time series.
                If meet error, then return None.
        """
        (
            sub_chunks,
            range_start,
            range_end,
            nbts,
            update_unix_time,
            is_last_element,
            success,
        ) = ctl_next_network_chunk(self._chunk_timeline_id)

        if not success:
            return None

        return (
            ChunkData(sub_chunks, nbts, Range(range_start, range_end)),
            update_unix_time,
            is_last_element,
        )

    def jump_to_and_get_disk_chunk(self, index: int) -> tuple[ChunkData, int] | None:
        """
        jump_to_and_get_disk_chunk moves to a specific time point of this timeline who is in index.
        Note that the returned ChunkData is in disk encoding.

        jump_to_and_get_disk_chunk is a very useful replacement of next_disk_chunk when you are trying
        to jump to a specific time point and no need to get the information of other time point.

        Note that if jump_to_and_get_disk_chunk return None (meet error), then the underlying pointer will
        back to the firest time point due to when an error occurs, some of the underlying data maybe is
        inconsistent.

        Time complexity: O(4096×n + C×(d+1)).
            - n is the sub chunk count of this chunk.
            - d is the distance between index and current pointer.
            - C is relevant to the average changes of all these time point.

        Args:
            index (int): The index of target time point that you want to jump to.

        Returns:
            tuple[ChunkData, int] | None:
                The chunk data of target time point.
                Returned int is the update unix time of this time point.
                If meet error, then return None.
        """
        sub_chunks, range_start, range_end, nbts, update_unix_time, _, success = (
            ctl_jump_to_disk_chunk(self._chunk_timeline_id, index)
        )
        if not success:
            return None
        return (
            ChunkData(sub_chunks, nbts, Range(range_start, range_end)),
            update_unix_time,
        )

    def jump_to_and_get_network_chunk(self, index: int) -> tuple[ChunkData, int] | None:
        """
        jump_to_and_get_disk_chunk moves to a specific time point of this timeline who is in index.
        Note that the returned ChunkData is in network encoding.

        jump_to_and_get_disk_chunk is a very useful replacement of next_disk_chunk when you are trying
        to jump to a specific time point and no need to get the information of other time point.

        Note that if jump_to_and_get_disk_chunk return None (meet error), then the underlying pointer will
        back to the firest time point due to when an error occurs, some of the underlying data maybe is
        inconsistent.

        Time complexity: O(4096×n + C×(d+1)).
            - n is the sub chunk count of this chunk.
            - d is the distance between index and current pointer.
            - C is relevant to the average changes of all these time point.

        Args:
            index (int): The index of target time point that you want to jump to.

        Returns:
            tuple[ChunkData, int] | None:
                The chunk data of target time point.
                Returned int is the update unix time of this time point.
                If meet error, then return None.
        """
        sub_chunks, range_start, range_end, nbts, update_unix_time, _, success = (
            ctl_jump_to_network_chunk(self._chunk_timeline_id, index)
        )
        if not success:
            return None
        return (
            ChunkData(sub_chunks, nbts, Range(range_start, range_end)),
            update_unix_time,
        )

    def last_disk_chunk(self) -> tuple[ChunkData, int] | None:
        """
        last_disk_chunk gets the latest time point
        of current chunk and the NBT blocks in it.

        Time complexity: Time complexity: O(4096×n).
        n is the sub chunk count of this chunk.

        Returns:
            tuple[ChunkData, int, bool] | None:
                The chunk data who is encoded in disk encoding.
                Returned int is the update unix time of the time point.
                If meet error, then return None.
        """
        sub_chunks, range_start, range_end, nbts, update_unix_time, success = (
            ctl_last_disk_chunk(self._chunk_timeline_id)
        )
        if not success:
            return None
        return (
            ChunkData(sub_chunks, nbts, Range(range_start, range_end)),
            update_unix_time,
        )

    def last_network_chunk(self) -> tuple[ChunkData, int] | None:
        """
        last_network_chunk gets the latest time point
        of current chunk and the NBT blocks in it.

        Time complexity: O(4096×n).
        n is the sub chunk count of this chunk.

        Returns:
            tuple[ChunkData, int, bool] | None:
                The chunk data who is encoded in network encoding.
                Returned int is the update unix time of the time point.
                If meet error, then return None.
        """
        sub_chunks, range_start, range_end, nbts, update_unix_time, success = (
            ctl_last_network_chunk(self._chunk_timeline_id)
        )
        if not success:
            return None
        return (
            ChunkData(sub_chunks, nbts, Range(range_start, range_end)),
            update_unix_time,
        )

    def pop(self):
        """
        pop tries to delete the first time point from this timeline.

        If current timeline is empty, or it is read only, or there is
        only one time point, then we will do no operation.

        Raises:
            Exception: When failed to pop.
        """
        err = ctl_pop(self._chunk_timeline_id)
        if len(err) > 0:
            raise Exception(err)

    def save(self):
        """
        save saves current timeline into the underlying database, and also release current timeline.

        Read only timeline should also calling save to release the resource.
        But read only timeline calling this function will only release but don't do further operation.
        Additionally, empty non read only timeline is also follow the same behavior.

        Note that you could use s.Empty() and s.ReadOnly() to check.

        If you calling save with no exception, then this timeline is released and can't be used again.
        Also, you can't call save multiple times.
        But, if exception happened, then this object will not released.

        Note that we will not check whether it has been released, nor will we check whether you have called
        save multiple times.

        save must calling at the last modification of the timeline; otherwise, the timeline will not be able
        to maintain data consistency.

        Raises:
            Exception: When failed to save this timeline.
        """
        err = ctl_save(self._chunk_timeline_id)
        if len(err) > 0:
            raise Exception(err)



from .types import LIB
from .types import as_c_string, as_python_string
from .types import CInt, CLongLong, CString


LIB.NewTimelineDB.argtypes = [CString, CInt, CInt]
LIB.ReleaseTimelineDB.argtypes = [CLongLong]
LIB.CloseTimelineDB.argtypes = [CLongLong]
LIB.NewChunkTimeline.argtypes = [CLongLong, CInt, CInt, CInt, CInt]
LIB.ReleaseChunkTimeline.argtypes = [CLongLong]
LIB.DeleteChunkTimeline.argtypes = [CLongLong, CInt, CInt, CInt]
LIB.LoadLatestTimePointUnixTime.argtypes = [CLongLong, CInt, CInt, CInt]
LIB.SaveLatestTimePointUnixTime.argtypes = [CLongLong, CInt, CInt, CInt, CLongLong]

LIB.NewTimelineDB.restype = CLongLong
LIB.ReleaseTimelineDB.restype = None
LIB.CloseTimelineDB.restype = CString
LIB.NewChunkTimeline.restype = CLongLong
LIB.ReleaseChunkTimeline.restype = None
LIB.DeleteChunkTimeline.restype = CString
LIB.LoadLatestTimePointUnixTime.restype = CLongLong
LIB.SaveLatestTimePointUnixTime.restype = CString


def new_timeline_db(path: str, no_grow_sync: bool, no_sync: bool) -> int:
    return int(LIB.NewTimelineDB(as_c_string(path), CInt(no_grow_sync), CInt(no_sync)))


def release_timeline_db(id: int) -> None:
    LIB.ReleaseTimelineDB(CLongLong(id))


def tldb_close_timeline_db(id: int) -> str:
    return as_python_string(LIB.CloseTimelineDB(CLongLong(id)))


def tldb_new_chunk_timeline(
    id: int, dm: int, posx: int, posz: int, read_only: bool
) -> int:
    return int(
        LIB.NewChunkTimeline(
            CLongLong(id), CInt(dm), CInt(posx), CInt(posz), CInt(read_only)
        )
    )


def release_chunk_timeline(id: int) -> None:
    LIB.ReleaseChunkTimeline(CLongLong(id))


def tldb_delete_chunk_timeline(id: int, dm: int, posx: int, posz: int) -> str:
    return as_python_string(
        LIB.DeleteChunkTimeline(CLongLong(id), CInt(dm), CInt(posx), CInt(posz))
    )


def tldb_load_latest_time_point_unix_time(
    id: int, dm: int, posx: int, posz: int
) -> int:
    return int(
        LIB.LoadLatestTimePointUnixTime(CLongLong(id), CInt(dm), CInt(posx), CInt(posz))
    )


def tldb_save_latest_time_point_unix_time(
    id: int, dm: int, posx: int, posz: int, time_stamp: int
) -> str:
    return as_python_string(
        LIB.SaveLatestTimePointUnixTime(
            CLongLong(id), CInt(dm), CInt(posx), CInt(posz), CLongLong(time_stamp)
        )
    )



from .define import Dimension, ChunkPos
from .constant import DIMENSION_OVERWORLD
from dataclasses import dataclass
from .chunk_timeline import ChunkTimeline
from ..internal.symbol_export_timeline_db import (
    new_timeline_db,
    release_timeline_db,
    tldb_close_timeline_db,
    tldb_delete_chunk_timeline,
    tldb_load_latest_time_point_unix_time,
    tldb_new_chunk_timeline,
    tldb_save_latest_time_point_unix_time,
)


@dataclass
class TimelineDatabase:
    """
    TimelineDatabase wrapper and implements all features from Timeline,
    and as a provider to provide timeline of chunk related functions.
    """

    _database_id: int = -1

    def __del__(self):
        if self._database_id >= 0 and release_timeline_db is not None:
            release_timeline_db(self._database_id)

    def is_valid(self) -> bool:
        """
        is_valid check current timeline database is valid or not.

        If not valid, it means the timeline database actually not exist,
        not only Python but also in Go.

        Try to use an invalid timeline database is not allowed,
        and any operation will be terminated.

        Returns:
            bool: Whether the timeline database is valid or not.
        """
        return self._database_id >= 0

    def close_timeline_db(self):
        """
        close_timeline_db closes the timeline database.

        It will wait until all the timelines in use are
        released before closing the database.

        Raises:
            Exception: When failed to close the timeline database.
        """
        err = tldb_close_timeline_db(self._database_id)
        if len(err) > 0:
            raise Exception(err)

    def new_chunk_timeline(
        self,
        pos: ChunkPos,
        read_only: bool = False,
        dm: Dimension = DIMENSION_OVERWORLD,
    ) -> ChunkTimeline:
        """
        new_chunk_timeline gets the timeline of a chunk who is at pos.
        If current timeline database is not exist or meet error, the you get a invalid ChunkTimeline.
        Therefore, you need use ChunkTimeline.is_valid() to check whether the timeline you get is valid or not.

        Note that if timeline of current chunk is not exist, then we will not create a timeline
        but return an empty one so you can modify it. The time to create the timeline is only when you
        save a timeline that not empty to the database.

        If read_only is true, then returned a timeline but only can read.
        For a read only timeline, you also need use ChunkTimeline.save to release it.

        Important:
            - Once any modifications have been made to the returned timeline, you must save them
              at the end; otherwise, the timeline will not be able to maintain data consistency
              (only need to save at the last modification).

            - Timeline of one chunk can't be using by multiple threads. Therefore, you will
              get blocking when a thread calling new_chunk_timeline but there is still some
              threads are using target chunk.

            - Calling ChunkTimeline.save to release the timeline.

            - Returned ChunkTimeline can't shared with multiple threads, and it's your responsibility
              to ensure this thing.

        Args:
            pos (ChunkPos): The chunk position of the target chunk.
            read_only (bool, optional): You want to the target timeline is read only or not.
                                        Defaults to False.
            dm (Dimension, optional): The dimension of the target chunk.
                                      Defaults to DIMENSION_OVERWORLD.
        """
        return ChunkTimeline(
            tldb_new_chunk_timeline(self._database_id, int(dm), pos.x, pos.z, read_only)
        )

    def delete_chunk_timeline(self, pos: ChunkPos, dm: Dimension = DIMENSION_OVERWORLD):
        """
        delete_chunk_timeline deletes the timeline of chunk who at pos.
        If timeline is not exist, then do no operation.

        Time complexity: O(n).
        n is the time point that this chunk have.

        Args:
            pos (ChunkPos): The chunk position of the target chunk.
            dm (Dimension, optional): The dimension of the target chunk.
                                      Defaults to DIMENSION_OVERWORLD.

        Raises:
            Exception: When failed to delete target timeline.
        """
        err = tldb_delete_chunk_timeline(self._database_id, int(dm), pos.x, pos.z)
        if len(err) > 0:
            raise Exception(err)

    def load_latest_time_point_unix_time(
        self, pos: ChunkPos, dm: Dimension = DIMENSION_OVERWORLD
    ):
        """
        load_latest_time_point_unix_time loads
        the time when latest time point update.

        If not exist, then return 0.

        Args:
            pos (ChunkPos): The chunk position of the target chunk.
            dm (Dimension, optional): The dimension of the target chunk.
                                      Defaults to DIMENSION_OVERWORLD.

        Returns:
            int: The unix time of the latest time point.
                 Return 0 for not exist or current timeline database is not exist.
        """
        result = tldb_load_latest_time_point_unix_time(
            self._database_id, int(dm), pos.x, pos.z
        )
        if result == -1:
            return 0
        return result

    def save_latest_time_point_unix_time(
        self, pos: ChunkPos, time_stamp: int, dm: Dimension = DIMENSION_OVERWORLD
    ):
        """
        save_latest_time_point_unix_timesaves the time
        when the latest time point is generated.

        If timeStamp is 0, then delete the time from the database.

        Args:
            pos (ChunkPos): The chunk position of the target chunk.
            time_stamp (int): The unix time to update.
            dm (Dimension, optional): The dimension of the target chunk.
                                      Defaults to DIMENSION_OVERWORLD.

        Raises:
            Exception: When failed to update the unix time.
        """
        err = tldb_save_latest_time_point_unix_time(
            self._database_id, int(dm), pos.x, pos.z, time_stamp
        )
        if len(err) > 0:
            raise Exception(err)


def new_timeline_database(
    path: str, no_grow_sync: bool = False, no_sync: bool = False
) -> TimelineDatabase:
    """
    new_timeline_database open a level database that used for
    chunk delta update whose at path.

    If not exist, then create a new database.

    Note that you could use TimelineDatabase.is_valid() to check
    whether the timeline database is valid or not.

    Args:
        path (str): The path of the timeline database want to open or create.
        no_grow_sync (bool, optional): When no_grow_sync is true, skips the truncate call when growing the database.
                             Setting this to true is only safe on non-ext3/ext4 systems.
                             Skipping truncation avoids preallocation of hard drive space and
                             bypasses a truncate() and fsync() syscall on remapping.
                                - See also: https://github.com/boltdb/bolt/issues/284
                            Defaults to False.
        no_sync (bool, optional): Setting the no_sync flag will cause the database to skip fsync()
                        calls after each commit. This can be useful when bulk loading data
                        into a database and you can restart the bulk load in the event of
                        a system failure or database corruption. Do not set this flag for
                        normal use.
                        THIS IS UNSAFE. PLEASE USE WITH CAUTION.
                        Defaults to False.

    Returns:
        TimelineDatabase: The opened timeline database.
    """
    return TimelineDatabase(new_timeline_db(path, no_grow_sync, no_sync))



from .define import Dimension, Range

DIMENSION_ID_OVERWORLD = 0
DIMENSION_ID_NETHER = 1
DIMENSION_ID_END = 2

DIMENSION_OVERWORLD = Dimension(DIMENSION_ID_OVERWORLD)
DIMENSION_NETHER = Dimension(DIMENSION_ID_NETHER)
DIMENSION_END = Dimension(DIMENSION_ID_END)

RANGE_OVERWORLD = DIMENSION_OVERWORLD.range()
RANGE_NETHER = DIMENSION_NETHER.range()
RANGE_END = DIMENSION_END.range()
RANGE_INVALID = Range(0, -1)



from .timeline.define import Range, Dimension, ChunkPos
from .timeline.constant import RANGE_OVERWORLD, RANGE_NETHER, RANGE_END
from .timeline.constant import DIMENSION_OVERWORLD, DIMENSION_NETHER, DIMENSION_END

from .timeline.define import ChunkData
from .timeline.timeline_database import new_timeline_database



from dataclasses import dataclass, field


@dataclass(frozen=True)
class Range:
    """
    Range represents the height range of a Dimension in blocks.

    The first value of the Range holds the minimum Y value,
    the second value holds the maximum Y value.

    Note that Range is a hashable and cannot be further modified object.
    """

    start_range: int = 0
    end_range: int = 0


@dataclass(frozen=True)
class Dimension:
    """
    Dimension is a dimension of a World.

    It influences a variety of properties of a World such as the building range,
    the sky color and the behavior of liquid blocks.

    Note that Dimension is a hashable and cannot be further modified object.

    Args:
        dm (int): The id of this dimension.
    """

    dm: int = 0

    def __int__(self) -> int:
        return self.dm

    def range(self) -> Range:
        """range returns the range that player could build block in this dimension.

        Returns:
            Range: The range that player could build block in this dimension.
                   If this dimension is not standard dimension, then redirect
                   to Overworld range.
        """
        match self.dm:
            case 0:
                return Range(-64, 319)
            case 1:
                return Range(0, 127)
            case 2:
                return Range(0, 255)
            case _:
                return Range(-64, 319)

    def height(self) -> int:
        """
        height returns the height of this dimension.

        For example, the height of Overworld is 384
        due to "384 = 319 - (-64) + 1", and 319 is
        the max Y that Overworld could build, and -64
        is the min Y that Overworld could build.

        Returns:
            int: The height of this dimension.
                 If this dimension is not standard dimension, then redirect
                 to Overworld height.
        """
        match self.dm:
            case 0:
                return 384
            case 1:
                return 128
            case 2:
                return 256
            case _:
                return 384

    def __str__(self) -> str:
        match self.dm:
            case 0:
                return "Overworld"
            case 1:
                return "Nether"
            case 2:
                return "End"
            case _:
                return f"Custom (id={self.dm})"


@dataclass(frozen=True)
class ChunkPos:
    """
    ChunkPos holds the position of a chunk. The type is provided as a utility struct for keeping track of a
    chunk's position.

    Chunks do not themselves keep track of that. Chunk positions are different from block positions in the
    way that increasing the X/Z by one means increasing the absolute value on the X/Z axis in terms of blocks
    by 16.

    Note that ChunkPos is a hashable and cannot be further modified object.
    """

    x: int = 0
    z: int = 0


@dataclass
class ChunkData:
    """
    ChunkData represents the data for a chunk.
    A single chunk could holds the block matrix data and its block entities NBT data.

    Note that for nbts field in this class, we recommend the length of nbts is equal
    to the block entities in this chunk.
    That means, each element in this list is represents the NBT data of one block entity.
    For example, if this chunk has T NBT blocks, then len(nbts) will be T.

    Note that the length of nbts for you is not strict, you can combine multiple block
    entities payload to just one element, this will not case problems.
    However, we granted that the length of the nbts list you get by calling the function
    (next_disk_chunk, next_network_chunk, last_disk_chunk and last_network_chunk) must be
    the number of block entities within this chunk, so that each element in this list is
    just one little endian TAG_Compound NBT.

    Args:
        sub_chunks (list[bytes]): The payload (block matrix data) of this chunk.
                                  The length of this list must equal to 24 if this chunk is from Overworld,
                                  or 8 if this chunk is from Nether, or 16 if this chunk is from End.
                                  For example, a Overworld chunk have 24 sub chunks, and sub_chunks list is
                                  holds all the sub chunk data for this chunk, so len(sub_chunks) is 24.
        nbts: (list[bytes]): The block entities NBT data of this chunk.
        chunk_range: (Range, optional):
            The range of this chunk.
            For a Overworld chunk, this is Range(-64, 319);
            for a Nether chunk, this is Range(0, 127);
            for a End chunk, this is Range(0, 255).
            Defaults to Range(-64, 319).
    """

    sub_chunks: list[bytes] = field(default_factory=lambda: [])
    nbts: list[bytes] = field(default_factory=lambda: [])
    chunk_range: Range = Range(-64, 319)
